package com.jiuqi.mde;

import java.util.List;
import com.jiuqi.util.DuplicationException;
import java.util.ArrayList;

public class MDERecordSet
{
    protected MDEDataInfo dataInfo;
    protected MDEFieldSet fields;
    protected boolean hasOrder;
    protected ArrayList records;
    protected MDEIndex dimValuesIndex;
    protected ArrayList additionalIndexes;
    private boolean isUseLoadDataSync;
    protected Object mutexLoadData;
    protected Object mutexBuildIndex;
    
    public MDERecordSet(final MDEDataInfo dataInfo) {
        this.dataInfo = null;
        this.fields = null;
        this.hasOrder = false;
        this.records = new ArrayList(40);
        this.dimValuesIndex = null;
        this.additionalIndexes = new ArrayList();
        this.isUseLoadDataSync = false;
        this.mutexLoadData = new Object();
        this.mutexBuildIndex = new Object();
        this.dataInfo = dataInfo;
        this.fields = dataInfo.fields();
        this.hasOrder = dataInfo.dimensions().hasOrder();
    }
    
    public void setIsUseLoadDataSync(final boolean flag) {
        this.isUseLoadDataSync = flag;
    }
    
    public boolean getIsUseLoadDataSync() {
        return this.isUseLoadDataSync;
    }
    
    public final String getName() {
        return this.dataInfo.getName();
    }
    
    public final MDEDataInfo getDataInfo() {
        return this.dataInfo;
    }
    
    public final int size() {
        return this.records.size();
    }
    
    public final MDERecord get(final int index) {
        return this.records.get(index);
    }
    
    public final void add(final MDERecord record) throws DuplicationException, MDEException {
        if (this.dimValuesIndex != null && (record.dimValues != null || !record.isInserted())) {
            if (this.hasOrder) {
                this.dimValuesIndex.genOrder(record.getDimValues());
            }
            this.addToDVIndex(record);
        }
        this.records.add(record);
    }
    
    public final void remove(final int index) {
        final MDERecord record = this.records.get(index);
        if (this.dimValuesIndex != null && record.dimValues != null) {
            this.dimValuesIndex.remove((Object)record.getDimValues());
        }
        record.emptyRecord();
        this.records.remove(index);
    }
    
    public final void remove(final MDERecord record) {
        if (this.dimValuesIndex != null && record.dimValues != null) {
            this.dimValuesIndex.remove((Object)record.getDimValues());
        }
        record.emptyRecord();
        this.records.remove(record);
        if (this.additionalIndexes != null) {
            this.additionalIndexes.clear();
        }
    }
    
    public final void removeAll(final List removeRecords) {
        for (int i = 0; i < removeRecords.size(); ++i) {
            this.remove(removeRecords.get(i));
        }
    }
    
    protected final void addToDVIndex(final MDERecord record) throws DuplicationException, MDEException {
        record.refreshDimValues(null);
        if (!this.dataInfo.isUnique()) {
            this.dimValuesIndex.addlast((Object)record.getDimValues(), (Object)record);
        }
        else {
            try {
                this.dimValuesIndex.genOrder(record.getDimValues());
                this.dimValuesIndex.add((Object)record.getDimValues(), (Object)record);
            }
            catch (final DuplicationException ex) {
                if (this.hasOrder) {
                    record.getDimValues().setValue("顺序号", null);
                    this.dimValuesIndex.genOrder(record.getDimValues());
                    this.dimValuesIndex.add((Object)record.getDimValues(), (Object)record);
                    return;
                }
                throw ex;
            }
        }
    }
    
    protected final void buildDimValuesIndex() throws MDEException {
        if (this.getIsUseLoadDataSync()) {
            synchronized (this.mutexBuildIndex) {
                if (this.dimValuesIndex == null) {
                    this.buildDimValuesIndex0();
                }
                monitorexit(this.mutexBuildIndex);
                return;
            }
        }
        this.buildDimValuesIndex0();
    }
    
    private void buildDimValuesIndex0() throws MDEException {
        if (this.dataInfo.dimensions().size() == 0) {
            throw new MDEException("没有定义维度，无法使用键值查找功能");
        }
        if (this.dimValuesIndex == null) {
            this.dimValuesIndex = new MDEIndex();
        }
        else {
            this.dimValuesIndex.clear();
        }
        try {
            for (int i = 0, count = this.records.size(); i < count; ++i) {
                final MDERecord record = this.records.get(i);
                this.addToDVIndex(record);
            }
        }
        catch (final DuplicationException ex) {
            throw new MDEException((Exception)ex);
        }
    }
    
    protected final MDERecord dvFind(final MDEDimValueSet dimValues) throws MDEException {
        if (this.dimValuesIndex == null) {
            this.buildDimValuesIndex();
        }
        return (MDERecord)this.dimValuesIndex.findkey((Object)dimValues);
    }
    
    protected final MDERecord dvFindNoOrder(final MDEDimValueSet dimValues) throws MDEException {
        if (this.dimValuesIndex == null) {
            this.buildDimValuesIndex();
        }
        if (this.hasOrder) {
            final int index = this.dimValuesIndex.findNoOrder(dimValues);
            return (MDERecord)this.dimValuesIndex.getValue(index);
        }
        return (MDERecord)this.dimValuesIndex.findkey((Object)dimValues);
    }
    
    protected final List dvFindMatches(final MDEDimValueSet speciafication) throws MDEException {
        if (this.dimValuesIndex == null) {
            this.buildDimValuesIndex();
        }
        List result = null;
        if (this.getIsUseLoadDataSync()) {
            synchronized (this.dimValuesIndex) {
                result = this.dvFindMatches0(speciafication);
                monitorexit(this.dimValuesIndex);
                return result;
            }
        }
        result = this.dvFindMatches0(speciafication);
        return result;
    }
    
    private List dvFindMatches0(final MDEDimValueSet speciafication) {
        List result = null;
        int start = 0;
        int stop = this.dimValuesIndex.size();
        boolean allSpeciafied = false;
        if (speciafication.dimensions().equals(this.dataInfo.dimensions()) && speciafication.getValue(0) != null) {
            int index = -1;
            for (int i = 0, count = speciafication.size(); i < count; ++i) {
                if (speciafication.getValue(i) == null) {
                    index = i;
                    break;
                }
            }
            allSpeciafied = (index < 0);
            if (allSpeciafied) {
                start = this.dimValuesIndex.searchfirst((Object)speciafication);
                if (start < 0) {
                    start = stop;
                }
                else {
                    stop = this.dimValuesIndex.searchlast((Object)speciafication);
                    if (stop < this.dimValuesIndex.size()) {
                        ++stop;
                    }
                }
            }
            else {
                final MDEDimension dimension = speciafication.dimensions().get(index);
                final MDEDimValueSet find = (MDEDimValueSet)speciafication.clone();
                find.setValue(index, dimension.getMinValue());
                start = this.dimValuesIndex.searchfirst((Object)find);
                if (start < 0) {
                    start = -1 - start;
                }
                find.setValue(index, dimension.getMaxValue());
                stop = this.dimValuesIndex.searchlast((Object)find);
                if (stop < 0) {
                    stop = -1 - stop;
                }
                else {
                    ++stop;
                }
                if (stop > this.dimValuesIndex.size()) {
                    --stop;
                }
            }
        }
        result = new ArrayList(stop - start);
        for (int j = start; j < stop; ++j) {
            final MDERecord record = (MDERecord)this.dimValuesIndex.getValue(j);
            if (allSpeciafied || record.getDimValues().matches(speciafication)) {
                result.add(record);
            }
        }
        return result;
    }
    
    protected final void rebuildFieldIndex(final MDEAdditionalIndex index) {
        index.clear();
        index.setCapacity(this.records.size());
        final MDEField field = index.getKeyFields().get(0);
        for (int i = 0, count = this.records.size(); i < count; ++i) {
            final MDERecord record = this.records.get(i);
            index.addlast(record.getFieldValue(field), (Object)record);
        }
    }
    
    protected MDEAdditionalIndex createAdditionalIndex(final MDEField field) {
        final MDEAdditionalIndex result = new MDEAdditionalIndex();
        result.setField(field);
        return result;
    }
    
    protected final MDEAdditionalIndex getAdditionalIndex(final MDEField field) {
        MDEAdditionalIndex result = null;
        for (int i = 0, count = this.additionalIndexes.size(); i < count; ++i) {
            final MDEAdditionalIndex index = this.additionalIndexes.get(i);
            if (index.keyFieldsMatch(field)) {
                result = index;
                break;
            }
        }
        if (result == null) {
            result = this.createAdditionalIndex(field);
            this.rebuildFieldIndex(result);
            this.additionalIndexes.add(result);
        }
        return result;
    }
    
    protected final MDERecord fvFind(final MDEField field, final Object fieldValue) {
        final MDEAdditionalIndex index = this.getAdditionalIndex(field);
        final MDERecord rec = (MDERecord)index.findkey(fieldValue);
        return rec;
    }
    
    public void resetRecords() {
        if (this.additionalIndexes != null) {
            this.additionalIndexes.clear();
        }
        if (this.dimValuesIndex != null) {
            this.dimValuesIndex.clear();
            this.dimValuesIndex = null;
        }
        this.records.clear();
    }
}
